from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import DefendSummonBase

from ...modifiable_values import CombatActionValue, DamageIncreaseValue
from ...event import RoundPrepareEventArguments

from ...action import Actions
from ...struct import Cost

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    PlayerActionLabels,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    PassiveSkillBase,
    CharacterBase,
    SkillTalent,
)


class Reflection_3_3(DefendSummonBase):
    name: Literal["Reflection"] = "Reflection"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    damage: int = 1
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1
    attack_until_run_out_of_usage: bool = False


class MirrorReflectionOfDoom(ElementalSkillBase):
    name: Literal["Mirror Reflection of Doom"] = "Mirror Reflection of Doom"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO,
        elemental_dice_number=3,
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [self.create_summon("Reflection")]


class StellarisPhantasm(ElementalBurstBase):
    name: Literal["Stellaris Phantasm"] = "Stellaris Phantasm"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO,
        elemental_dice_number=3,
        charge=3,
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [self.create_team_status("Illusory Bubble")]


class IllusoryTorrent(PassiveSkillBase):
    name: Literal["Illusory Torrent"] = "Illusory Torrent"
    usage: int = 1

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        When in round prepare, reset usage
        """
        self.usage = 1
        return []

    def value_modifier_COMBAT_ACTION(
        self,
        value: CombatActionValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> CombatActionValue:
        """
        When combat action is switch, player index is self, switch from mona to
        other people, has usage, and currently is a combat action,
        change the combat action to quick action and decrease usage.
        """
        if value.action_label & PlayerActionLabels.SWITCH.value == 0:
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
        ):
            return value
        if self.usage <= 0:
            return value
        # as mona is active character, current switch is always combat action
        assert value.do_combat_action
        value.do_combat_action = False
        assert mode == "REAL"
        self.usage -= 1
        return value


class ProphecyOfSubmersion_3_3(SkillTalent):
    name: Literal["Prophecy of Submersion"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Mona"] = "Mona"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO,
        elemental_dice_number=3,
        charge=3,
    )
    skill: Literal["Stellaris Phantasm"] = "Stellaris Phantasm"

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        """
        If mona is active character, and damage triggered hydro reaction,
        which is made by self, increase damage by 2.
        """
        if value.damage_type != DamageType.DAMAGE:
            # not damage, not modify
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            source_area=ObjectPositionType.CHARACTER,  # quipped
            source_is_active_character=True,  # active character
            player_idx_same=True,  # self damage
        ):
            return value
        if ElementType.HYDRO not in value.reacted_elements:
            # no hydro reaction, not activate
            return value
        value.damage += 2
        return value


class Mona_3_3(CharacterBase):
    name: Literal["Mona"]
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        ElementalNormalAttackBase
        | MirrorReflectionOfDoom
        | StellarisPhantasm
        | IllusoryTorrent
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT,
    ]
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Ripple of Fate",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            MirrorReflectionOfDoom(),
            StellarisPhantasm(),
            IllusoryTorrent(),
        ]


register_class(Mona_3_3 | ProphecyOfSubmersion_3_3 | Reflection_3_3)



from ..utils import BaseModel, list_unique_range_right
from typing import Literal, List, Any
from .consts import DieColor
from .struct import Cost, MultipleObjectPosition, ObjectPosition


class RequestBase(BaseModel):
    """
    Base class of request.
    """

    name: Literal["RequestBase"] = "RequestBase"
    player_idx: int


class SwitchCardRequest(RequestBase):
    name: Literal["SwitchCardRequest"] = "SwitchCardRequest"
    card_names: List[str]
    maximum_switch_number: int = 999


class ChooseCharacterRequest(RequestBase):
    name: Literal["ChooseCharacterRequest"] = "ChooseCharacterRequest"
    available_character_idxs: List[int]


class RerollDiceRequest(RequestBase):
    """
    Request for reroll dice.

    Args:
        colors (List[DieColor]): The colors of the dice to reroll.
        reroll_times (int): The times of reroll. In one time, player can choose
            any number of dice to reroll, and after one reroll, the reroll
            times will decrease by 1. If it reaches 0, the request is removed.
    """

    name: Literal["RerollDiceRequest"] = "RerollDiceRequest"
    colors: List[DieColor]
    reroll_times: int


class SwitchCharacterRequest(RequestBase):
    """
    Request for switch character. It can be combat or quick request.
    """

    name: Literal["SwitchCharacterRequest"] = "SwitchCharacterRequest"
    active_character_idx: int
    target_character_idx: int
    dice_colors: List[DieColor]
    cost: Cost


class ElementalTuningRequest(RequestBase):
    """
    Request for elemental tuning.
    """

    name: Literal["ElementalTuningRequest"] = "ElementalTuningRequest"
    dice_colors: List[DieColor]
    dice_idxs: List[int]
    card_idxs: List[int]


class DeclareRoundEndRequest(RequestBase):
    """
    Request for declare round end.
    """

    name: Literal["DeclareRoundEndRequest"] = "DeclareRoundEndRequest"


class UseSkillRequest(RequestBase):
    """
    Request for use skill.
    """

    name: Literal["UseSkillRequest"] = "UseSkillRequest"
    character_idx: int
    skill_idx: int
    dice_colors: List[DieColor]
    cost: Cost


class UseCardRequest(RequestBase):
    """
    Request for use card.
    """

    name: Literal["UseCardRequest"] = "UseCardRequest"
    card_idx: int
    dice_colors: List[DieColor]
    targets: List[ObjectPosition | MultipleObjectPosition]
    cost: Cost


class ResponseBase(BaseModel):
    """
    Base class of response.
    """

    name: Literal["ResponseBase"] = "ResponseBase"
    request: RequestBase

    @property
    def player_idx(self) -> int:
        """
        Return the player idx of the response.
        """
        return self.request.player_idx

    def is_valid(self, match: Any) -> bool:
        """
        Check whether the response is valid.
        """
        return True


class SwitchCardResponse(ResponseBase):
    """
    Card indices want to switch, based on the request.
    """

    name: Literal["SwitchCardResponse"] = "SwitchCardResponse"
    request: SwitchCardRequest
    card_idxs: List[int]

    def is_valid(self, match: Any) -> bool:
        if len(self.card_idxs) > self.request.maximum_switch_number:
            return False
        return list_unique_range_right(
            self.card_idxs, minn=0, maxn=len(self.request.card_names)
        )

    @property
    def card_names(self) -> List[str]:
        """
        Return the card names of the response.
        """
        return [self.request.card_names[i] for i in self.card_idxs]


class ChooseCharacterResponse(ResponseBase):
    name: Literal["ChooseCharacterResponse"] = "ChooseCharacterResponse"
    request: ChooseCharacterRequest
    character_idx: int

    def is_valid(self, match: Any) -> bool:
        return self.character_idx in self.request.available_character_idxs


class RerollDiceResponse(ResponseBase):
    name: Literal["RerollDiceResponse"] = "RerollDiceResponse"
    request: RerollDiceRequest
    reroll_dice_idxs: List[int]

    def is_valid(self, match: Any) -> bool:
        """
        if have duplicate dice indices, or dice indices out of range,
        return False.
        """
        return list_unique_range_right(
            self.reroll_dice_idxs, minn=0, maxn=len(self.request.colors)
        )


class SwitchCharacterResponse(ResponseBase):
    name: Literal["SwitchCharacterResponse"] = "SwitchCharacterResponse"
    request: SwitchCharacterRequest
    dice_idxs: List[int]

    def is_valid(self, match: Any) -> bool:
        """
        Cost matches the request.
        """
        if not list_unique_range_right(
            self.dice_idxs, minn=0, maxn=len(self.request.dice_colors)
        ):
            return False
        cost_colors = [self.request.dice_colors[i] for i in self.dice_idxs]
        return self.request.cost.is_valid(cost_colors, 0, False)


class ElementalTuningResponse(ResponseBase):
    name: Literal["ElementalTuningResponse"] = "ElementalTuningResponse"
    request: ElementalTuningRequest
    dice_idx: int
    card_idx: int

    def is_valid(self, match: Any) -> bool:
        """
        Check whether the response is valid.
        """
        return (
            self.dice_idx in self.request.dice_idxs
            and self.card_idx in self.request.card_idxs
        )


class DeclareRoundEndResponse(ResponseBase):
    name: Literal["DeclareRoundEndResponse"] = "DeclareRoundEndResponse"
    request: DeclareRoundEndRequest


class UseSkillResponse(ResponseBase):
    """
    Response for use skill. Currently not support choose skill target.
    """

    name: Literal["UseSkillResponse"] = "UseSkillResponse"
    request: UseSkillRequest
    dice_idxs: List[int]

    def is_valid(self, match: Any) -> bool:
        """
        Check whether the response is valid.
        """
        if not list_unique_range_right(
            self.dice_idxs, minn=0, maxn=len(self.request.dice_colors)
        ):
            return False
        cost_colors = [self.request.dice_colors[i] for i in self.dice_idxs]
        table = match.player_tables[self.request.player_idx]
        charge, arcane_legend = table.get_charge_and_arcane_legend()
        return self.request.cost.is_valid(cost_colors, charge, arcane_legend)


class UseCardResponse(ResponseBase):
    name: Literal["UseCardResponse"] = "UseCardResponse"
    request: UseCardRequest
    dice_idxs: List[int]
    target: ObjectPosition | MultipleObjectPosition | None

    def is_valid(self, match: Any) -> bool:
        """
        Check whether the response is valid.
        """
        # dice color right
        if not list_unique_range_right(
            self.dice_idxs, minn=0, maxn=len(self.request.dice_colors)
        ):
            return False
        cost_colors = [self.request.dice_colors[i] for i in self.dice_idxs]
        table = match.player_tables[self.request.player_idx]
        charge, arcane_legend = table.get_charge_and_arcane_legend()
        if not self.request.cost.is_valid(cost_colors, charge, arcane_legend):
            return False
        # if has targets, response target should not be None
        if self.target is None and len(self.request.targets) > 0:
            return False
        # if response target is not None, it should be in targets
        if self.target is not None:
            if self.target not in self.request.targets:
                return False
        return True


Requests = (
    SwitchCardRequest
    | ChooseCharacterRequest
    | RerollDiceRequest
    | SwitchCharacterRequest
    | ElementalTuningRequest
    | DeclareRoundEndRequest
    | UseSkillRequest
    | UseCardRequest
)

Responses = (
    SwitchCardResponse
    | ChooseCharacterResponse
    | RerollDiceResponse
    | SwitchCharacterResponse
    | ElementalTuningResponse
    | DeclareRoundEndResponse
    | UseSkillResponse
    | UseCardResponse
)



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from .base import ArtifactBase

from ....consts import DieColor, ObjectPositionType

from ....struct import Cost, ObjectPosition

from ....event import CharacterDefeatedEventArguments
from ....action import CreateDiceAction, Actions


class GamblersEarrings_3_8(ArtifactBase):
    name: Literal["Gambler's Earrings"]
    version: Literal["3.8"] = "3.8"
    cost: Cost = Cost(same_dice_number=1)
    usage: int = 3

    def equip(self, match: Any) -> List[Actions]:
        """
        Equip this artifact. Reset usage.
        """
        self.usage = 3
        return []

    def event_handler_CHARACTER_DEFEATED(
        self, event: CharacterDefeatedEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        When an opposing character is defeated, check if the character this
        card is attached to is the active character. If so, create Omni
        Element x2.
        """
        target_position = ObjectPosition(
            player_idx=event.action.player_idx,
            # all the following are not used to check, no need to set correctly
            area=ObjectPositionType.INVALID,
            id=-1,
        )
        if self.position.not_satisfy(
            "both pidx=diff and source area=character active=true",
            target_position,
            match,
        ):
            # not opponent character defeated, or self not active, or self not equipped
            return []
        if self.usage <= 0:
            # no usage left
            return []
        self.usage -= 1
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx, number=2, color=DieColor.OMNI
            )
        ]


class GamblersEarrings_3_3(GamblersEarrings_3_8):
    version: Literal["3.3"]
    usage: int = 999

    def equip(self, match: Any) -> List[Actions]:
        """
        Equip this artifact. Reset usage.
        """
        self.usage = 999
        return []


register_class(GamblersEarrings_3_8 | GamblersEarrings_3_3)



from typing import Any, Dict, List, Literal

from .....utils.class_registry import register_class
from ....status.team_status.base import DefendTeamStatus
from ....summon.base import AttackAndGenerateStatusSummonBase, SwirlChangeSummonBase
from ....action import (
    Actions,
    CreateObjectAction,
    MakeDamageAction,
    RemoveObjectAction,
    SwitchCharacterAction,
)
from ....status.character_status.base import RoundEndAttackCharacterStatus
from ....match import Match
from ....event import (
    ReceiveDamageEventArguments,
    RoundEndEventArguments,
    RoundPrepareEventArguments,
)
from ....struct import Cost
from ....consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    IconType,
    ObjectPositionType,
    WeaponType,
)
from ....character.character_base import (
    CharacterBase,
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    SkillTalent,
)
from .....utils.desc_registry import DescDictType


class OverawingAssault_4_3(RoundEndAttackCharacterStatus):
    name: Literal["Overawing Assault"] = "Overawing Assault"
    version: Literal["4.3"] = "4.3"
    usage: int = 1
    max_usage: int = 1
    damage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.PIERCING
    icon_type: IconType = IconType.DEBUFF

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.hp >= 6:
            # attack self
            return list(super().event_handler_ROUND_END(event, match))
        else:
            # remove self
            self.usage = 0
            return list(self.check_should_remove())


class BogglecatBoxStatus_4_3(DefendTeamStatus):
    name: Literal["Bogglecat Box"] = "Bogglecat Box"
    version: Literal["4.3"] = "4.3"
    usage: int = 1
    max_usage: int = 1
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1


class BogglecatBox_4_3(SwirlChangeSummonBase, AttackAndGenerateStatusSummonBase):
    name: Literal["Bogglecat Box"] = "Bogglecat Box"
    version: Literal["4.3"] = "4.3"
    usage: int = 2
    max_usage: int = 2
    damage: int = 1

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Match
    ) -> List[Actions]:
        """
        When anyone receives damage, and is pyro/hydro/cryo/electro, and is
        our character, and current damage type is anemo, change damage element.
        """
        if self.damage_elemental_type != DamageElementalType.ANEMO:
            # already changed, do nothing
            return []
        if event.final_damage.damage_elemental_type not in [
            DamageElementalType.PYRO,
            DamageElementalType.HYDRO,
            DamageElementalType.CRYO,
            DamageElementalType.ELECTRO,
        ]:
            # damage element not pyro/hydro/cryo/electro, do nothing
            return []
        if event.final_damage.target_position.player_idx != self.position.player_idx:
            # not our player made damage, do nothing
            return []
        # do type change
        self.damage_elemental_type = event.final_damage.damage_elemental_type
        return []


class EnigmaticFeint(ElementalSkillBase):
    name: Literal["Enigmatic Feint"] = "Enigmatic Feint"
    cost: Cost = Cost(elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3)
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    use_time_this_round: int = 0

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        """
        reset use_time_this_round
        """
        self.use_time_this_round = 0
        return []

    def get_actions(self, match: Match) -> List[Actions]:
        self.use_time_this_round += 1
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        talent_equipped = character.talent is not None
        if talent_equipped and self.use_time_this_round == 2:
            # second use with talent, deal +2 damage
            self.damage += 2
        ret = super().get_actions(match)
        damage_action = ret[0]
        assert isinstance(damage_action, MakeDamageAction)
        if talent_equipped and self.use_time_this_round == 2:
            # second use with talent, reset damage and force opponent to switch
            self.damage -= 2
            prev_character_idx = match.player_tables[
                1 - self.position.player_idx
            ].previous_character_idx()
            if prev_character_idx is not None:
                return [
                    damage_action,
                    SwitchCharacterAction(
                        player_idx=1 - self.position.player_idx,
                        character_idx=prev_character_idx,
                    ),
                    self.charge_self(1),
                ]
        elif self.use_time_this_round == 1 and character.hp <= 8:
            # first use, heal 2 hp and attach status
            heal = self.attack_self(match, -2)
            damage_action.damage_value_list += heal.damage_value_list
            ret.append(
                CreateObjectAction(
                    object_name="Overawing Assault",
                    object_position=character.position.set_area(
                        ObjectPositionType.CHARACTER_STATUS
                    ),
                    object_arguments={},
                )
            )
        return ret


class MagicTrickAstonishingShift(ElementalBurstBase):
    name: Literal["Magic Trick: Astonishing Shift"] = "Magic Trick: Astonishing Shift"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3, charge=2
    )
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ANEMO

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(match) + [self.create_summon("Bogglecat Box", {})]


class AColdBladeLikeAShadow_4_3(SkillTalent):
    name: Literal["A Cold Blade Like a Shadow"]
    version: Literal["4.3"] = "4.3"
    character_name: Literal["Lynette"] = "Lynette"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO,
        elemental_dice_number=3,
    )
    skill: Literal["Enigmatic Feint"] = "Enigmatic Feint"


class Lynette_4_3(CharacterBase):
    name: Literal["Lynette"]
    version: Literal["4.3"] = "4.3"
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | EnigmaticFeint | MagicTrickAstonishingShift
    ] = []
    faction: List[FactionType] = [
        FactionType.FONTAINE,
        FactionType.FATUI,
        FactionType.ARKHE_OUSIA,
    ]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Rapid Ritesword",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            EnigmaticFeint(),
            MagicTrickAstonishingShift(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTER/Lynette": {
        "names": {"en-US": "Lynette", "zh-CN": "琳妮特"},
        "image_path": "cardface/Char_Avatar_Linette.png",  # noqa E501
        "id": 1508,
        "descs": {"4.3": {"en-US": "", "zh-CN": ""}},
    },
    "SKILL_Lynette_NORMAL_ATTACK/Rapid Ritesword": {
        "names": {"en-US": "Rapid Ritesword", "zh-CN": "迅捷礼刺剑"},
        "descs": {
            "4.3": {"en-US": "Deals 2 Physical DMG.", "zh-CN": "造成2点物理伤害。"}
        },
    },
    "SKILL_Lynette_ELEMENTAL_SKILL/Enigmatic Feint": {
        "names": {"en-US": "Enigmatic Feint", "zh-CN": "谜影障身法"},
        "descs": {
            "4.3": {
                "en-US": "Deals 3 Anemo DMG. The first time this Skill is used this Round, heal the user for 2 HP if they have no more than 8 HP, but apply Overawing Assault to the user.",  # noqa: E501
                "zh-CN": "造成3点风元素伤害，本回合第一次使用此技能、且自身生命值不多于8时治疗自身2点，但是附属攻袭余威。",  # noqa: E501
            }
        },
    },
    "CHARACTER_STATUS/Overawing Assault": {
        "names": {"en-US": "Overawing Assault", "zh-CN": "攻袭余威"},
        "descs": {
            "4.3": {
                "en-US": "End Phase: If this character's HP is at least 6, then they will take 2 Piercing DMG.\nDuration (Rounds): 1",  # noqa: E501
                "zh-CN": "结束阶段：如果角色生命值至少为6，则受到2点穿透伤害。\n持续回合：1",  # noqa: E501
            }
        },
    },
    "SKILL_Lynette_ELEMENTAL_BURST/Magic Trick: Astonishing Shift": {
        "names": {"en-US": "Magic Trick: Astonishing Shift", "zh-CN": "魔术·运变惊奇"},
        "descs": {
            "4.3": {
                "en-US": "Deals 2 Anemo DMG, summons 1 Bogglecat Box.",
                "zh-CN": "造成2点风元素伤害，召唤惊奇猫猫盒。",
            }
        },
    },
    "TEAM_STATUS/Bogglecat Box": {
        "names": {"en-US": "Bogglecat Box", "zh-CN": "惊奇猫猫盒"},
        "descs": {
            "4.3": {
                "en-US": "When your active character takes DMG: Decreases DMG taken by 1.",  # noqa: E501
                "zh-CN": "我方出战角色受到伤害时：抵消1点伤害。",
            }
        },
    },
    "SUMMON/Bogglecat Box": {
        "names": {"en-US": "Bogglecat Box", "zh-CN": "惊奇猫猫盒"},
        "descs": {
            "4.3": {
                "en-US": "End Phase: Deals 1 Anemo DMG.\nUsage(s): 2\n\nWhen your active character takes DMG: Decreases DMG taken by 1. (Once per Round)\nWhen your character(s) take Cryo/Hydro/Pyro/Electro DMG: Change the Elemental Type of this card. (Once before leaving play)",  # noqa: E501
                "zh-CN": "结束阶段：造成1点风元素伤害。\n可用次数：2\n\n我方出战角色受到伤害时：抵消1点伤害。（每回合1次）\n我方角色受到冰/水/火/雷伤害时：转换此牌的元素类型。（离场前仅限一次）",  # noqa: E501
            }
        },
        "image_path": "cardface/Summon_Linette.png",  # noqa: E501
    },
    "TALENT_Lynette/A Cold Blade Like a Shadow": {
        "names": {"en-US": "A Cold Blade Like a Shadow", "zh-CN": "如影流露的冷刃"},
        "descs": {
            "4.3": {
                "en-US": "Combat Action: When your active character is Lynette, equip this card.\nAfter Lynette equips this card, immediately use Enigmatic Feint once.\nAfter your Lynette, who has this card equipped, uses Enigmatic Feint for the second time this Round: deals +2 DMG and forces your opponent to switch to their previous character.\n(You must have Lynette in your deck to add this card to your deck.)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为琳妮特时，装备此牌。\n琳妮特装备此牌后，立刻使用一次谜影障身法。\n装备有此牌的琳妮特每回合第二次使用谜影障身法时：伤害+2，并强制敌方切换到前一个角色。\n（牌组中包含琳妮特，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Constellation_Linette.png",  # noqa: E501
        "id": 215081,
    },
}


register_class(
    OverawingAssault_4_3
    | BogglecatBoxStatus_4_3
    | BogglecatBox_4_3
    | AColdBladeLikeAShadow_4_3
    | Lynette_4_3,
    desc,
)



from lpsim.server.struct import ObjectPosition
from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp,
    check_usage,
    get_pidx_cidx,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)


def test_vanilla_weapons():
    # use frontend and FastAPI server to perform commands, and test commands
    # that start with TEST. NO NOT put TEST at the end of command list!
    # If a command is successfully performed, frontend will print history
    # commands in console. Note that frontend cannot distinguish if a new
    # match begins, so you need to refresh the page before recording a new
    # match, otherwise the history commands will be mixed.
    #
    # for tests, it starts with TEST and contains a test id, which is used to
    # identify the test. Other texts in the command are ignored, but you can
    # parse them if you want. Refer to the following code.
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "end",
            "end",
            "end",
            "TEST 2 weapon target right",
            "card 0 0 0 1",
            "skill 1 0 1 2",
            "TEST 1 10 10 90 10 10 8 10 90 10 10",
            "sw_char 2 0",
            "skill 1 0 1 2",
            "end",
            "TEST 1 10 10 85 10 10 8 10 90 10 9",
            "end",
            "card 1 0 0 1",
            "card 3 0 0 1",
            "sw_char 1 0",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "end",
            "end",
            "end",
            "TEST 1 10 10 90 10 10 8 10 90 10 10",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "TEST 1 10 10 89 10 10 8 10 90 10 10",
            "sw_char 1 0",
            "sw_char 0 0",
            "card 5 0 0 1",
            "skill 0 0 1",
            "TEST 1 10 10 89 10 10 8 10 90 10 10",
            "sw_char 4 0",
            "skill 1 0 1 2",
            "end",
            "sw_char 0 0",
            "skill 2 0 1 2",
            "TEST 1 10 10 78 10 10 8 10 90 10 9",
            "end",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Arataki Itto
        character:Barbara
        character:Noelle
        character:PhysicalMob
        character:Fischl
        Magic Guide*2
        Raven Bow*2
        Traveler's Handy Sword*2
        White Iron Greatsword*2
        White Tassel*2
        """
    )
    deck.characters[2].hp = 90
    deck.characters[2].max_hp = 90
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:5], hps[5:]]
                check_hp(match, hps)
            elif test_id == 2:
                cards = match.player_tables[0].hands
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        card = cards[req.card_idx]
                        targets = req.targets
                        for target in targets:
                            assert isinstance(target, ObjectPosition)
                            assert target.character_idx != 3  # 3 is mob
                        if card.name == "Magic Guide":
                            assert len(targets) == 1
                            assert isinstance(targets[0], ObjectPosition)
                            assert targets[0].character_idx == 1
                        elif card.name == "White Iron Greatsword":
                            assert len(targets) == 2
                            assert isinstance(targets[0], ObjectPosition)
                            assert isinstance(targets[1], ObjectPosition)
                            assert targets[0].character_idx == 0
                            assert targets[1].character_idx == 2
                        elif card.name == "Raven Bow":
                            assert len(targets) == 1
                            assert isinstance(targets[0], ObjectPosition)
                            assert targets[0].character_idx == 4
                        else:
                            raise AssertionError("Other weapon cannot use")
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_the_bell():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 0 1 0 1 2",
            "skill 0 0 1 2",
            "TEST 1 p1 status 1 usage 1",
            "end",
            "end",
            "end",
            "sw_char 1 0",
            "TEST 1 p1 status 1 usage 2",
            "skill 1 0 1 2",
            "TEST 1 p1 status 1 usage 2",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "TEST 1 p0 status 0 usage 0",
            "card 0 0 0 1 2",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "TEST 1 p1 status 1 usage 1",
            "end",
            "skill 1 0 1 2",
            "TEST 1 p1 status 1 usage 2",
            "end",
            "end",
            "skill 1 0 1 2",
            "card 0 0 0 1 2",
            "skill 1 0 1 2",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Arataki Itto
        character:Barbara
        character:Noelle
        The Bell*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cmd = cmd.strip().split()
                pid = int(cmd[2][1])
                snum = int(cmd[4])
                usage = int(cmd[6])
                status = match.player_tables[pid].team_status
                assert len(status) == snum
                if snum:
                    assert status[0].usage == usage
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_a_thousand():
    cmd_records = [
        [
            "sw_card",
            "choose 1",
            "skill 1 0 1 2",
            "sw_char 2 0",
            "skill 1 0 1 2",
            "end",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "sw_char 1 0",
            "sw_char 2 0",
            "choose 1",
            "sw_char 0 0",
            "sw_char 1 0",
            "TEST 1 3 7 0 10 10 4",
            "end",
        ],
        [
            "sw_card",
            "choose 2",
            "card 0 0 0 1 2",
            "skill 1 0",
            "skill 0 0 1 2",
            "sw_char 0 0",
            "skill 1 0 1 2",
            "TEST 1 10 10 6 10 10 6",
            "end",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck1 = Deck.from_str(
        """
        default_version:4.0
        character:Venti
        character:Xingqiu
        character:Noelle
        Mondstadt Hash Brown*30
        """
    )
    deck2 = Deck.from_str(
        """
        default_version:4.0
        character:Nahida
        character:Maguu Kenki
        character:Yoimiya
        A Thousand Floating Dreams*30
        """
    )
    match.set_deck([deck1, deck2])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_vortex_vanquisher():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 1 0 15 14 13",
            "skill 1 12 11 10",
            "sw_char 3 9",
            "skill 0 8 7 6",
            "sw_char 0 5",
            "skill 0 4 3 2",
            "TEST 1 8 10 10 10 10 10 10 2",
            "TEST 2 p0 usage 1",
            "end",
            "skill 1 15 14 13",
            "skill 1 12 11 10",
            "TEST 1 7 10 10 10 10 10 0 2",
            "end",
            "skill 1 15 14 13",
            "TEST 1 2 10 10 10 10 7 0 2",
            "sw_char 3 12",
            "skill 0 11 10 9",
            "sw_char 0 8",
            "skill 1 7 6 5",
            "end",
            "skill 1 15 14 13",
            "skill 0 12 11 10",
            "TEST 1 2 10 10 8 1 0 0 2",
            "end",
        ],
        [
            "sw_card 3 2",
            "choose 3",
            "skill 1 15 14 13",
            "TEST 1 10 10 10 10 10 10 10 6",
            "skill 1 12 11 10",
            "end",
            "sw_char 2 15",
            "skill 1 14 13 12",
            "TEST 1 8 10 10 10 10 10 5 2",
            "TEST 2 p0 usage 1",
            "sw_char 1 11",
            "sw_char 2 10",
            "choose 1",
            "TEST 1 8 10 10 10 10 10 0 2",
            "skill 0 9 8 7",
            "end",
            "skill 1 15 14 13",
            "skill 0 12 11 10",
            "sw_char 0 9",
            "sw_char 1 8",
            "sw_char 0 7",
            "end",
            "sw_char 1 15",
            "TEST 2 p0 usage 3",
            "end",
            "choose 0",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Candace*2
        character:Diluc
        character:Ningguang
        Sweet Madame*15
        Vortex Vanquisher*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:4], hps[4:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].team_status) == 1
                assert match.player_tables[pidx].team_status[0].usage == int(cmd[-1])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_vortex_2():
    cmd_records = [
        [
            "sw_card",
            "choose 3",
            "sw_char 2 15",
            "skill 1 14 13 12",
            "sw_char 3 11",
            "TEST 1 10 10 10 7 10 7 10 10",
            "card 3 2 10 9 8",
            "card 2 0",
            "skill 1 7 6",
            "skill 0 5 4 3",
        ],
        [
            "sw_card",
            "choose 2",
            "sw_char 1 15",
            "sw_char 2 14",
            "skill 1 13 12 11",
            "sw_char 1 10",
            "TEST 1 10 10 10 8 10 1 10 10",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Candace*2
        character:Chongyun
        character:Hu Tao
        Sweet Madame*15
        Vortex Vanquisher*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:4], hps[4:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_lithic_spear():
    cmd_records = [
        [
            "sw_card 1 3",
            "choose 1",
            "card 1 0 15 14 13",
            "TEST 2 p0c1 usage 3",
            "skill 1 12 11 10",
            "skill 0 9 8 7",
            "skill 0 6 5 4",
            "skill 1 3 2 1",
            "end",
            "card 1 0 15 14 13",
            "skill 0 12 11 10",
            "skill 0 9 8 7",
            "card 0 0",
            "card 4 0 6",
            "TEST 2 p0c1 usage 3",
            "skill 0 5 4 3",
            "skill 1 2 1 0",
            "end",
            "skill 1 15 14 13",
            "choose 4",
            "TEST 1 10 0 10 10 10 10 10 0 0 3",
            "end",
            "card 3 1 15 14 13",
            "skill 1 12 11 10",
            "skill 0 9 8 7",
            "skill 2 6 5 4",
            "skill 1 3 2 1",
            "end",
        ],
        [
            "sw_card 1 2 4",
            "choose 3",
            "TEST 2 p0c1 usage 3 2",
            "skill 1 15 14 13",
            "TEST 2 p0c1 usage",
            "TEST 1 10 10 10 10 10 10 10 10 6 10",
            "end",
            "choose 2",
            "skill 1 15 14",
            "sw_char 4 13",
            "skill 1 12 11 10",
            "sw_char 2 9",
            "skill 0 8 7 6",
            "choose 4",
            "card 0 1 5 4 3",
            "TEST 2 p1c4 usage 3",
            "skill 0 2 1 0",
            "end",
            "skill 2 15 14 13",
            "TEST 1 10 5 10 10 10 10 10 0 0 3",
            "card 0 1 12 11 10",
            "TEST 2 p1c4 usage 2",
            "skill 1 9 8 7",
            "end",
            "end",
            "choose 0",
            "choose 1",
            "card 4 0 15 14 13",
            "TEST 2 p1c1 usage",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Xingqiu
        character:Candace
        character:Hu Tao
        character:Chongyun
        character:Shenhe
        Lithic Spear*10
        Lithic Spear@3.3*10
        Where Is the Unseen Razor?*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:5], hps[5:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                cidx = int(cmd[2][3])
                usages = [int(x) for x in cmd[4:]]
                status = match.player_tables[pidx].characters[cidx].status
                assert len(usages) == len(status)
                for u, s in zip(usages, status):
                    assert u == s.usage
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_kings_squire():
    cmd_records = [
        [
            "sw_card 3 1 2",
            "choose 2",
            "card 2 1 15 14 13",
            "skill 0 12 11 10",
            "sw_char 0 9",
            "card 1 0",
            "card 1 0 8",
            "skill 1 7",
            "end",
        ],
        [
            "sw_card 1 2 3",
            "choose 0",
            "card 1 0 15 14 13",
            "skill 1 12",
            "sw_char 2 11",
            "card 0 0",
            "card 3 1 10",
            "card 1 0 9 8",
            "TEST 1 6 10 8 7 10 8",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Nahida
        character:Collei
        Where Is the Unseen Razor?*10
        King's Squire*10
        Floral Sidewinder*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_kings_squire_2():
    cmd_records = [
        [
            "sw_card 4 2",
            "choose 2",
            "card 1 0 15 14 13",
            "card 1 0",
            "card 3 1 12",
            "skill 1 11",
            "sw_char 1 10",
            "end",
        ],
        [
            "sw_card 2 1",
            "choose 0",
            "card 1 0 15 14 13",
            "card 2 0",
            "card 3 1 12",
            "sw_char 2 11",
            "skill 1 10",
            "TEST 2 card 0 cost 2",
            "TEST 2 card 1 cost 4",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Kamisato Ayaka
        character:Collei
        Where Is the Unseen Razor?*10
        King's Squire*10
        Floral Sidewinder*10
        Kanten Senmyou Blessing*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 2:
                cmd = cmd.split()
                cidx = int(cmd[3])
                cost = int(cmd[5])
                card = match.player_tables[1].hands[cidx]
                assert card.cost.total_dice_cost == cost
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_amos():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 0 0 15 14 13",
            "skill 1 12 11 10",
            "sw_char 1 9",
            "TEST 1 8 7 10 10 10 8",
            "card 0 0",
            "card 0 1 8",
            "skill 2 7 6 5 4 3",
            "TEST 1 8 2 10 7 7 2",
            "sw_char 2 2",
            "end",
            "sw_char 1 15",
            "skill 2 14 13 12 11 10",
            "TEST 1 8 2 8 4 4 0",
            "skill 2 9 8 7 6 5",
            "TEST 1 8 2 8 1 2 0",
            "end",
        ],
        [
            "sw_card",
            "choose 2",
            "skill 0 15 14 13",
            "card 0 2 12 11 10",
            "skill 0 9 8 7",
            "TEST 1 8 7 10 7 7 2",
            "skill 2 6 5 4",
            "sw_char 0 3",
            "sw_char 2 2",
            "end",
            "end",
            "choose 0",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Ganyu
        character:Collei
        Where Is the Unseen Razor?*10
        Amos' Bow*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_fruit_of_fullfillment():
    cmd_records = [
        [
            "sw_card",
            "choose 1",
            "card 2 2 15 14 13",
            "card 0 2 12 11 10",
            "card 2 0",
            "card 5 2 9",
            "TEST 1 p0 hand 8",
            "card 4 0",
            "card 4 2 8",
            "card 5 0",
            "card 4 2 7",
            "TEST 1 p0 hand 10",
            "card 4 2 6 5 4",
            "TEST 1 p0 hand 10",
            "end",
        ],
        ["sw_card", "choose 0", "card 0 0 15 14 13", "TEST 1 p1 hand 6", "end"],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Nahida
        character:Mona
        character:Klee
        Where Is the Unseen Razor?*10
        Fruit of Fulfillment*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                hand = match.player_tables[pidx].hands
                assert len(hand) == int(cmd[4])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_sacrificial():
    cmd_records = [
        [
            "sw_card 2 3",
            "choose 0",
            "card 1 0 15 14 13",
            "sw_char 1 12",
            "skill 1 11 10 9",
            "TEST 2 p1 dice 9",
            "skill 1 9 8 7",
            "TEST 2 p1 dice 7",
            "end",
            "skill 1 15 14 13",
            "skill 2 13 12 11",
            "TEST 1 9 2 9 2 8 4",
            "end",
        ],
        [
            "sw_card 1 2",
            "choose 1",
            "card 2 0 15 14 13",
            "sw_char 2 12",
            "TEST 2 p0 dice 10",
            "skill 0 11 10 9",
            "TEST 2 p0 dice 7",
            "skill 1 8 7 6",
            "card 3 0",
            "card 1 0 6",
            "sw_char 0 5",
            "skill 1 4 3 2",
            "TEST 2 p1 dice 3",
            "end",
            "TEST 2 p0 dice 14",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Chongyun
        character:Fischl
        character:Yae Miko
        Where Is the Unseen Razor?*10
        Sacrificial Greatsword*10
        Sacrificial Sword*10
        Sacrificial Bow*10
        Sacrificial Fragments*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                num = int(cmd[-1])
                assert len(match.player_tables[pidx].dice.colors) == num
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_wolf_gravestone():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 0 0 15 14 13",
            "skill 1 12 11 10",
            "skill 1 9 8 7",
            "sw_char 1 6",
            "end",
        ],
        [
            "sw_card",
            "choose 1",
            "skill 1 15 14 13",
            "choose 0",
            "card 0 0 12 11 10",
            "skill 1 9 8 7",
            "skill 0 6 5 4",
            "skill 0 3 2 1",
            "TEST 1 4 1 9 10 0 10",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Chongyun
        character:Fischl
        character:Yae Miko
        Where Is the Unseen Razor?*10
        Wolf's Gravestone*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_aquila_favonia():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 0 15 14 13",
            "sw_char 1 12",
            "skill 0 11 10 9",
            "sw_char 0 8",
            "card 0 0 7 6 5",
            "skill 1 4 3 2",
            "end",
            "TEST 1 7 7 9 8 8 7",
            "skill 2 15 14 13",
            "TEST 1 8 7 9 8 8 5",
            "skill 0 12 11 10",
            "skill 0 9 8 7",
            "TEST 1 2 6 8 8 7 0",
            "sw_char 1 6",
            "end",
            "sw_char 0 15",
            "choose 2",
            "TEST 1 0 2 8 8 9 0",
            "sw_char 1 0",
            "skill 0 0 1 2",
            "end",
        ],
        [
            "sw_card",
            "choose 1",
            "skill 0 15 14 13",
            "sw_char 0 12",
            "skill 0 11 10 9",
            "sw_char 2 8",
            "end",
            "skill 0 15 14 13",
            "skill 1 12 11 10",
            "skill 2 9 8 7",
            "choose 1",
            "skill 1 6 5 4",
            "end",
            "card 1 1 15 14 13",
            "skill 0 12 11 10",
            "skill 0 9 8 7",
            "sw_char 0 0",
            "TEST 1 0 2 8 5 9 0",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Xingqiu
        character:Qiqi
        character:Yae Miko
        Aquila Favonia*20
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_skyward():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 0 0 15 14 13",
            "skill 1 12 11 10",
            "TEST 1 10 10 10 10 10 7",
            "skill 0 9 8 7",
            "skill 0 6 5 4",
        ],
        [
            "sw_card",
            "choose 2",
            "card 0 0 15 14 13",
            "skill 1 12 11 10",
            "TEST 1 10 10 10 10 10 3",
            "sw_char 1 9",
            "TEST 1 10 10 10 10 7 3",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Xingqiu
        character:Kamisato Ayaka
        character:Yae Miko
        Skyward Atlas*3
        Skyward Blade*3
        Skyward Harp*3
        Skyward Pride*3
        Skyward Spine*3
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_favonius_lighting_elegy():
    cmd_records = [
        [
            "sw_card 1 2",
            "choose 0",
            "TEST 2 p0c1 charge 0",
            "TEST 2 p1c1 charge 0",
            "card 0 0 15 14 13",
            "card 8 0 12 11 10",
            "TEST 2 p0c1 charge 1",
            "sw_char 2 9",
            "skill 0 8 7 6",
            "sw_char 0 5",
            "TEST 2 p1c1 charge 1",
            "card 14 0",
            "card 3 0 4 3 2",
            "end",
            "card 1 0",
            "card 4 0 15",
            "TEST 2 p0c1 charge 1",
            "sw_char 1 14",
            "skill 1 13 12 11",
            "skill 2 10 9 8 7",
            "sw_char 0 6",
            "TEST 2 p1c2 charge 1",
            "TEST 1 21 17 30 30 27 25",
            "skill 2 5 4 3",
            "skill 1 2 1 0",
            "TEST 1 19 17 30 24 25 18",
            "end",
            "TEST 3 p0 usage 1 2",
            "TEST 3 p1 usage 3 2",
            "skill 0 15 14 13",
        ],
        [
            "sw_card 1 2 4 3",
            "choose 1",
            "skill 1 15 14 13",
            "card 0 0 12 11 10",
            "TEST 2 p1c1 charge 1",
            "skill 1 9 8 7",
            "skill 2 6 5 4 3",
            "card 13 0 2 1 0",
            "end",
            "sw_char 2 15",
            "skill 1 14 13 12",
            "TEST 2 p0c1 charge 0",
            "skill 2 11 10 9",
            "skill 1 8 7 6",
            "sw_char 0 5",
            "TEST 1 21 17 30 24 25 18",
            "skill 1 4 3 2",
            "end",
            "TEST 1 17 17 30 14 25 18",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Venti
        character:Xiangling
        character:Xingqiu
        Sweet Madame*15
        Elegy for the End*5
        Engulfing Lightning*5
        Favonius Sword*5
        Calx's Arts*5
        """
    )
    for character in deck.characters:
        character.max_hp = 30
        character.hp = 30
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    match.config.initial_hand_size = 20
    match.config.max_hand_size = 30
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.split()
                pidx, cidx = get_pidx_cidx(cmd)
                charge = int(cmd[-1])
                assert match.player_tables[pidx].characters[cidx].charge == charge
            elif test_id == 3:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                status = match.player_tables[pidx].team_status
                check_usage(status, cmd[4:])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    test_vanilla_weapons()
    # test_the_bell()
    # test_vortex_vanquisher()
    # test_vortex_2()
    # test_lithic_spear()
    # test_kings_squire()
    # test_kings_squire_2()
    # test_amos()
    # test_fruit_of_fullfillment()
    # test_sacrificial()
    # test_wolf_gravestone()
    # test_aquila_favonia()
    # test_skyward()
    test_favonius_lighting_elegy()



"""
Convert between deck codes and deck strings.
"""
import base64
from typing import List
import random
import json


# load name map and forbid list
deck_code_data = json.load(
    open(__file__.replace("deck_code.py", "deck_code_data.json"))
)
name_map = deck_code_data["name_map"][:]
forbid_list = deck_code_data["forbid_list"][:]
characters_idx = []
for i in range(len(name_map)):
    if name_map[i].startswith("character:"):
        characters_idx.append(i)
        name_map[i] = name_map[i][10:]


# create forbid word trie
class TrieNode:
    def __init__(self):
        self.children = {}
        self.is_end_of_word = False


class Trie:
    def __init__(self):
        self.root = TrieNode()

    def insert(self, word):
        node = self.root
        for char in word:
            if char not in node.children:
                node.children[char] = TrieNode()
            node = node.children[char]
        node.is_end_of_word = True

    def match(self, word):
        """
        test if word start with any word in trie
        """
        node = self.root
        for char in word:
            if char not in node.children:
                return False
            node = node.children[char]
            if node.is_end_of_word:
                return True
        return False

    def search(self, word):
        """
        test if word contains any word in trie
        """
        for i in range(len(word)):
            if self.match(word[i:]):
                return True
        return False


forbidden_trie = Trie()
for forbid in forbid_list:
    forbidden_trie.insert(forbid)


def deck_code_to_deck_str(deck_code: str, version: str | None = None) -> str:
    """
    Convert the base64 deck code to deck str. If version is set, add
    default_version.
    """
    binary = base64.b64decode(deck_code)
    bb = []
    for b in binary:
        bb.append((256 + b - binary[-1]) % 256)
    bb = bb[:-1]
    binary = bb[::2] + bb[1::2]
    res = ""
    for i in binary:
        res += "{:08b} ".format(i)
    # print(res)
    res = res.replace(" ", "")
    decode = []
    for i in range(0, len(res), 12):
        # print(res[i:i + 12], end = ' ')
        decode.append(int(res[i : i + 12], 2))
    decode = decode[:-1]
    results = []
    if version is not None:
        results.append(f"default_version:{version}")
    for x in decode:
        if x > 0 and x <= len(name_map):
            if x - 1 in characters_idx:
                results.append(f"character:{name_map[x - 1]}")
            else:
                results.append(name_map[x - 1])
    return "\n".join(results)


def _deck_str_to_deck_code_one(name_list: List[str], checksum: int) -> str:
    """
    Convert deck str to deck code. Versions about deck_str will be ignored.
    """
    binary = ""
    for i in name_list:
        if i not in name_map:
            # not found, use 0
            binary += "{:012b} ".format(0)
        else:
            idx = name_map.index(i)
            binary += "{:012b} ".format(idx + 1)
    binary = binary.replace(" ", "")
    b8 = []
    for i in range(0, len(binary), 8):
        b8.append(int(binary[i : i + 8], 2))
    b8[-1] = b8[-1] * 16  # 4 zeros
    uint = list(zip(b8[:25], b8[25:]))
    uint = [list(x) for x in uint]
    uint = sum(uint, start=[])
    uint.append(0)
    uint = [(x + checksum) % 256 for x in uint]
    res = base64.b64encode(bytes(uint)).decode()
    return res


def deck_str_to_deck_code(deck_str: str, max_retry_time: int = 10000) -> str:
    """
    convert deck str to deck code. Versions about deck_str will be ignored.
    As generated deck code may illegal, retry for max_retry_time times.
    If retry_time exceeded, raise ValueError.
    """
    deck_str_list = deck_str.strip().split("\n")
    deck_str_list = [x.strip() for x in deck_str_list]
    # remove empty line and comment
    deck_str_list = [x for x in deck_str_list if x != "" and x[0] != "#"]
    # remove default version
    deck_str_list = [x for x in deck_str_list if not x.startswith("default_version:")]
    # remove version mark
    deck_str_list = [x.split("@")[0] for x in deck_str_list]
    character_str_l = [x[10:] for x in deck_str_list if x.startswith("character:")]
    card_str_l = [x for x in deck_str_list if not x.startswith("character:")]
    character_str: List[str] = []
    card_str: List[str] = []
    for i in character_str_l:
        number = 1
        if "*" in i:
            i, number_str = i.split("*")
            number = int(number_str)
        character_str += [i] * number
    for i in card_str_l:
        number = 1
        if "*" in i:
            i, number_str = i.split("*")
            number = int(number_str)
        card_str += [i] * number
    if len(character_str) > 3 or len(card_str) > 30:
        raise ValueError("too many characters or cards")
    rand_checksum = list(range(256))
    random.shuffle(rand_checksum)
    for i in range(max_retry_time):
        # generate random checksum and shuffle cards
        if i > 255:
            checksum = random.randint(0, 255)
            random.shuffle(card_str)
        else:
            checksum = rand_checksum[i]
        # fill empty with ''
        name_list = (
            character_str
            + [""] * (3 - len(character_str))
            + card_str
            + [""] * (30 - len(card_str))
        )
        deck_code = _deck_str_to_deck_code_one(name_list, checksum)
        code_lower = deck_code.lower().replace("+", "")
        if not forbidden_trie.search(code_lower):
            return deck_code
    raise ValueError("in generating deck code: retry time exceeded")


__all__ = ["deck_code_to_deck_str", "deck_str_to_deck_code", "deck_code_data"]


if __name__ == "__main__":
    # descs = json.load(open(__file__.replace('deck_code.py',
    #                                         'default_desc.json'),
    #                        encoding = 'utf8'))
    # map_eng = [''] * len(map_chinese)
    # for key, value in descs.items():
    #     if 'names' not in value:
    #         continue
    #     name_c = value['names']['zh-CN']
    #     if name_c in map_chinese:
    #         idx = map_chinese.index(name_c)
    #         map_eng[idx] = value['names']["en-US"]
    # print(map_eng)

    # import json
    # res = json.dumps({
    #     'name_map': name_map,
    #     'forbid_list': forbid_list,
    # }, indent = 2)
    # open('deck_code_data.json', 'w', encoding = 'utf8').write(res)

    ...



import bisect
import logging
from typing import Any, Dict, get_args, get_type_hints

from .desc_registry import desc_exist, update_cost

from ..server.consts import ObjectType


def check_cls_valid_and_update_cost(cls: Any, obj_type: ObjectType):
    """
    Check whether this class hint and desc is valid for given class, if not, raise
    error. Then register cost information for this class.

    When registering cost, it initializes the class instance and get the cost from it.

    TODO: This function is added for compatibility with old logic, and is subject to
    refactor.
    """
    class_name = cls.__name__
    cls_type_hints = get_type_hints(cls)
    name_hints = cls_type_hints["name"]
    version_hints = cls_type_hints["version"]
    desc_hints = cls_type_hints["desc"]
    # check desc hint exist
    if len(get_args(desc_hints)) == 0:
        raise AssertionError(f"Class {cls} desc type hint is empty or not a Literal")
    descs = get_args(desc_hints)
    version = version_hints.__args__[0]
    type_hints = cls_type_hints["type"]
    if len(type_hints.__args__) != 1:
        raise AssertionError(f"Class {cls} type hint is empty or more than 1")
    obj_type = type_hints.__args__[0]
    names = name_hints.__args__
    # check if class name is match
    version_from_name = ".".join(class_name.split("_")[-2:])
    if version_from_name != version:
        raise AssertionError(
            f"Class {cls} version {version} is not equal to version "
            f"from name {version_from_name}"
        )
    for name in names:
        # check if an object can generate without name. No objects should be
        # generated without name.
        # TODO: is it necessary to distinguish?
        # direct_generate = False
        # try:
        #     instance = cls(version = version)
        #     direct_generate = True
        # except ValidationError:
        #     pass
        # if direct_generate and obj_type != ObjectType.SYSTEM:
        #     print(
        #         f"Class {cls} name {name} version {version} "
        #         "failed direct generate test"
        #     )
        desc_type = obj_type
        if obj_type == ObjectType.CHARACTER:
            # for character, also check its skill descs
            instance = cls(name=name, version=version)
            for skill in instance.skills:
                skill_type = f"SKILL_{name}_{skill.skill_type}"
                if not desc_exist(skill_type, skill.name, version):
                    raise AssertionError(
                        f"Class {cls} name {name} version {version} skill "
                        f"{skill.name} is not found in descs"
                    )
                # register skill costs
                update_cost(skill_type, skill.name, version, skill.cost)
        if obj_type == ObjectType.TALENT:
            # for talent, its type name should change
            target_character_names = cls_type_hints["character_name"].__args__
            assert len(target_character_names) == 1
            desc_type = f"TALENT_{target_character_names[0]}"
        for desc in descs:
            if desc == "":
                desc_name = name
            else:
                desc_name = f"{name}_{desc}"
            if desc == "":
                # empty, no extra desc
                if not desc_exist(desc_type, desc_name, version):
                    raise AssertionError(
                        f"Class {cls} name {name} version {version} type "
                        f"{desc_type} is not found in descs"
                    )
            else:
                if not desc_exist(desc_type, desc_name, version):
                    raise AssertionError(
                        f"Class {cls} name {name}(desc: {desc}) version "
                        f"{version} is not found in descs"
                    )
            if "cost" in cls_type_hints:
                # has cost, register it
                instance = cls(
                    name=name,
                    version=version,
                    position={"player_idx": -1, "area": "INVALID", "id": -1},
                )
                update_cost(desc_type, desc_name, version, instance.cost)


class InstanceFactory:
    instance_register = {}

    def __init__(self):
        self._instance_list = []
        self._instance_list_sorted = False

    def register_instance(self, cls: Any):
        """
        Register one class. If the class is registered, print a warning. If same
        version, name and base class, raise error.
        """
        cls_type_hints = get_type_hints(cls)
        version = cls_type_hints["version"].__args__[0]
        obj_type = cls_type_hints["type"].__args__[0]
        names = cls_type_hints["name"].__args__
        for name in names:
            check_cls_valid_and_update_cost(cls, obj_type)
            key = "%s+%s+%s" % (obj_type, name, version)
            if key in self.instance_register:
                if self.instance_register[key] == cls:
                    logging.warning(
                        f"Class {cls} is already registered in instance factory"
                    )
                else:
                    raise AssertionError(
                        f"Class {cls} and {self.instance_register[key]} have "
                        "same version and name"
                    )
            self.instance_register[key] = cls
            self._instance_list.append(key)
            self._instance_list_sorted = False

    def get_instance(self, base_class: Any, args: Dict):
        cls_type_hints = get_type_hints(base_class)
        obj_types = cls_type_hints["type"].__args__
        name = args["name"]
        version = args.get("version", "99.9")
        keys = []
        for obj_type in obj_types:
            keys.append("%s+%s+%s" % (obj_type, name, version))
        if not self._instance_list_sorted:
            self._instance_list.sort()
            self._instance_list_sorted = True
        instance_list = self._instance_list
        for key in keys:
            cls_key_idx = bisect.bisect_right(instance_list, key)
            if cls_key_idx == len(instance_list):
                cls_key_idx -= 1
            if instance_list[cls_key_idx] > key:
                cls_key_idx -= 1
            cls_key = instance_list[cls_key_idx]
            if self._is_same_instance(key, cls_key):
                cls = self.instance_register[cls_key]
                exact_version = cls_key.split("+")[2]
                args["version"] = exact_version
                return cls(**args)
            else:
                continue

        raise KeyError("Instance %s+%s not found in instance factory" % (name, version))

    @staticmethod
    def _is_same_instance(key1, key2):
        """
        Check whether type and name are the same.
        """
        info1 = key1.split("+")
        info2 = key2.split("+")
        if info1[0] == info2[0] and info1[1] == info2[1]:
            return True
        else:
            return False



from typing import Dict, List, Literal

from ....match import Match
from ....character import CharacterBase
from ....action import (
    ActionTypes,
    Actions,
    ChangeObjectUsageAction,
    DrawCardAction,
    MakeDamageAction,
)
from ....character.character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    SkillTalent,
)
from ....consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    ElementalReactionType,
    FactionType,
    IconType,
    ObjectPositionType,
    WeaponType,
)
from ....event import (
    ReceiveDamageEventArguments,
    RoundEndEventArguments,
    RoundPrepareEventArguments,
)
from ....modifiable_values import DamageValue
from ....status.character_status.base import PrepareCharacterStatus
from ....struct import Cost
from ....summon.base import AttackerSummonBase
from .....utils import register_class, DescDictType


class MujiMujiDaruma_4_4(AttackerSummonBase):
    name: Literal["Muji-Muji Daruma"] = "Muji-Muji Daruma"
    version: Literal["4.4"] = "4.4"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.ANEMO
    damage: int = 1
    heal: int = 2

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Match
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        ret = super().event_handler_ROUND_END(event, match)
        table = match.player_tables[self.position.player_idx]
        characters = table.characters
        # select character with most damage taken
        selected_character = characters[table.active_character_idx]
        for c in characters:
            if c.is_alive and c.damage_taken > selected_character.damage_taken:
                selected_character = c
        # heal character, and do default actions
        assert ret[0].type == ActionTypes.MAKE_DAMAGE
        ret[0].damage_value_list.append(
            DamageValue(
                position=self.position,
                damage_type=DamageType.HEAL,
                target_position=selected_character.position,
                damage=-self.heal,
                damage_elemental_type=DamageElementalType.HEAL,
                cost=Cost(),
            )
        )
        return ret


class FuufuuWhirlwindKickStatus_4_4(PrepareCharacterStatus):
    name: Literal["Fuufuu Whirlwind Kick"] = "Fuufuu Whirlwind Kick"
    version: Literal["4.4"] = "4.4"
    character_name: Literal["Sayu"] = "Sayu"
    skill_name: Literal["Fuufuu Whirlwind Kick"] = "Fuufuu Whirlwind Kick"
    element: DamageElementalType = DamageElementalType.ANEMO
    icon_type: IconType = IconType.SPECIAL

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Match
    ) -> List[Actions]:
        # if self make swirl, record swirled element
        if self.element != DamageElementalType.ANEMO:
            # not anemo, do nothing
            return []
        if not self.position.check_position_valid(
            event.final_damage.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not self use skill, skip
            return []
        if (
            event.final_damage.damage_from_element_reaction
            or event.final_damage.element_reaction != ElementalReactionType.SWIRL
        ):
            # not swirl or from elemental reaction, skip
            return []
        # record swirled element
        elements = event.final_damage.reacted_elements
        assert elements[0] == ElementType.ANEMO
        self.element = ELEMENT_TO_DAMAGE_TYPE[elements[1]]
        return []


class YooHooArtFuuinDash(ElementalSkillBase):
    name: Literal["Yoohoo Art: Fuuin Dash"] = "Yoohoo Art: Fuuin Dash"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3)

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_character_status("Fuufuu Whirlwind Kick"),
        ]


class FuufuuWhirlwindKick(ElementalSkillBase):
    name: Literal["Fuufuu Whirlwind Kick"] = "Fuufuu Whirlwind Kick"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost()

    def is_valid(self, match: Match) -> bool:
        return False

    def get_actions(self, match: Match) -> List[Actions]:
        # first change self element based on status, then call super
        status = (
            match.player_tables[self.position.player_idx]
            .characters[self.position.character_idx]
            .status
        )
        for s in status:
            if s.name == "Fuufuu Whirlwind Kick":
                self.damage_type = s.element  # type: ignore
                break
        else:
            raise AssertionError("No Fuufuu Whirlwind Kick status")
        return [self.attack_opposite_active(match, self.damage, self.damage_type)]


class YoohooArtMujinaFlurry(ElementalBurstBase):
    name: Literal["Yoohoo Art: Mujina Flurry"] = "Yoohoo Art: Mujina Flurry"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_summon("Muji-Muji Daruma"),
        ]


class SkivingNewAndImproved_4_4(SkillTalent):
    name: Literal["Skiving: New and Improved"]
    version: Literal["4.4"] = "4.4"
    cost: Cost = Cost(elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3)
    skill: Literal["Yoohoo Art: Fuuin Dash"] = "Yoohoo Art: Fuuin Dash"
    character_name: Literal["Sayu"] = "Sayu"
    usage: int = 1
    max_usage_per_round: int = 1

    def equip(self, match: Match) -> List[Actions]:
        self.usage = self.max_usage_per_round
        return []

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        self.usage = self.max_usage_per_round
        return []

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Match
    ) -> List[DrawCardAction]:
        if self.usage <= 0:
            # already used
            return []
        if not self.position.check_position_valid(
            event.final_damage.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
            source_area=ObjectPositionType.CHARACTER,
            source_is_active_character=True,
        ):
            # not self skill or not equipped or self not active
            return []
        if event.final_damage.element_reaction != ElementalReactionType.SWIRL:
            # not swirl
            return []
        # draw card
        self.usage -= 1
        return [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=2,
                draw_if_filtered_not_enough=True,
            )
        ]


class Sayu_4_4(CharacterBase):
    name: Literal["Sayu"]
    version: Literal["4.4"] = "4.4"
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase
        | YooHooArtFuuinDash
        | FuufuuWhirlwindKick
        | YoohooArtMujinaFlurry
    ] = []
    faction: List[FactionType] = [FactionType.INAZUMA]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Shuumatsuban Ninja Blade",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            YooHooArtFuuinDash(),
            FuufuuWhirlwindKick(),
            YoohooArtMujinaFlurry(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTER/Sayu": {
        "names": {"en-US": "Sayu", "zh-CN": "早柚"},
        "descs": {"4.4": {"en-US": "", "zh-CN": ""}},
        "image_path": "cardface/Char_Avatar_Sayu.png",  # noqa
        "id": 1507,
    },
    "SKILL_Sayu_NORMAL_ATTACK/Shuumatsuban Ninja Blade": {
        "names": {"en-US": "Shuumatsuban Ninja Blade", "zh-CN": "忍刀·终末番"},
        "descs": {
            "4.4": {"en-US": "Deals 2 Physical DMG.", "zh-CN": "造成2点物理伤害。"}
        },
    },
    "SKILL_Sayu_ELEMENTAL_SKILL/Yoohoo Art: Fuuin Dash": {
        "names": {"en-US": "Yoohoo Art: Fuuin Dash", "zh-CN": "呜呼流·风隐急进"},
        "descs": {
            "4.4": {
                "en-US": "Deals 1 Anemo DMG, and this character Prepare Skill|s1:prepares: Fuufuu Whirlwind Kick.\nIf this usage of the skill triggers Swirl, then Fuufuu Whirlwind Kick is converted to DMG of the Swirled Element.",  # noqa
                "zh-CN": "造成1点风元素伤害，本角色准备技能：风风轮舞踢。\n如果当前技能引发了扩散，则风风轮舞踢将改为造成被扩散元素的伤害。",  # noqa
            }
        },
    },
    "CHARACTER_STATUS/Fuufuu Whirlwind Kick": {
        "names": {"en-US": "Fuufuu Whirlwind Kick", "zh-CN": "风风轮舞踢"},
        "descs": {
            "4.4": {
                "en-US": "Elemental Skill\n(Prepare for 1 turn)\nDeals 2 Anemo DMG (or the Swirled Element's DMG)",  # noqa
                "zh-CN": "元素战技\n（需准备1个行动轮）\n造成2点风元素伤害（或被扩散元素的伤害）。",  # noqa
            }
        },
    },
    "SKILL_Sayu_ELEMENTAL_BURST/Yoohoo Art: Mujina Flurry": {
        "names": {"en-US": "Yoohoo Art: Mujina Flurry", "zh-CN": "呜呼流·影貉缭乱"},
        "descs": {
            "4.4": {
                "en-US": "Deals 1 Anemo DMG, summons 1 Muji-Muji Daruma.",
                "zh-CN": "造成1点风元素伤害，召唤不倒貉貉。",
            }
        },
    },
    "SUMMON/Muji-Muji Daruma": {
        "names": {"en-US": "Muji-Muji Daruma", "zh-CN": "不倒貉貉"},
        "descs": {
            "4.4": {
                "en-US": "End Phase: Deal 1 Anemo DMG, heal the character on your team that has taken the most damage for 2 HP.\nUsage(s): 2",  # noqa
                "zh-CN": "结束阶段：造成1点风元素伤害，治疗我方受伤最多的角色2点。\n可用次数：2",  # noqa
            }
        },
        "image_path": "cardface/Summon_Sayu.png",
    },
    "SKILL_Sayu_ELEMENTAL_SKILL/Fuufuu Whirlwind Kick": {
        "names": {"en-US": "Fuufuu Whirlwind Kick", "zh-CN": "风风轮舞踢"},
        "descs": {
            "4.4": {
                "en-US": "(Prepare for 1 turn)\nDeals 2 Anemo DMG (or the Swirled Element's DMG)",  # noqa
                "zh-CN": "（需准备1个行动轮）\n造成2点风元素伤害（或被扩散元素的伤害）。",  # noqa
            }
        },
    },
    "TALENT_Sayu/Skiving: New and Improved": {
        "names": {"en-US": "Skiving: New and Improved", "zh-CN": "偷懒的新方法"},
        "descs": {
            "4.4": {
                "en-US": "Combat Action: When your active character is Sayu, equip this card.\nWhen Sayu equips this card, immediately use Yoohoo Art: Fuuin Dash once.\nWhen your Sayu who has this card equipped is the active character, draw 2 cards when you trigger a Swirl reaction. (Once per Round)\n(You must have Sayu in your deck to add this card to your deck.)",  # noqa
                "zh-CN": "战斗行动：我方出战角色为早柚时，装备此牌。\n早柚装备此牌后，立刻使用一次呜呼流·风隐急进。\n装备有此牌的早柚为出战角色期间，我方引发扩散反应时：抓2张牌。（每回合1次）\n（牌组中包含早柚，才能加入牌组）",  # noqa
            }
        },
        "image_path": "cardface/Modify_Constellation_Sayu.png",  # noqa
        "id": 215071,
    },
}


register_class(
    Sayu_4_4
    | MujiMujiDaruma_4_4
    | FuufuuWhirlwindKickStatus_4_4
    | SkivingNewAndImproved_4_4,
    desc,
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...event import RoundEndEventArguments

from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageIncreaseValue, DamageValue

from ...action import (
    Actions,
    ChangeObjectUsageAction,
    MakeDamageAction,
    SwitchCharacterAction,
)
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class DandelionField_3_3(AttackerSummonBase):
    name: Literal["Dandelion Field"] = "Dandelion Field"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.ANEMO
    damage: int = 2

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        """
        If found Jean with talent, increase anemo damage by 1
        """
        if (
            value.damage_type != DamageType.DAMAGE
            or value.damage_elemental_type != DamageElementalType.ANEMO
            or value.position.player_idx != self.position.player_idx
            or value.target_position.player_idx == self.position.player_idx
        ):
            # not anemo damage, or not from our side attack enemy
            return value
        # find jean with talent
        characters = match.player_tables[self.position.player_idx].characters
        for character in characters:
            if character.name == "Jean" and character.talent is not None:
                # found, increase damage
                assert mode == "REAL"
                value.damage += 1
                return value
        return value

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        ret = super().event_handler_ROUND_END(event, match)
        our_active = match.player_tables[
            self.position.player_idx
        ].get_active_character()
        ret.append(
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=our_active.position,
                        damage=-1,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=Cost(),
                    )
                ],
            )
        )
        return ret


# Skills


class GaleBlade(ElementalSkillBase):
    name: Literal["Gale Blade"] = "Gale Blade"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        next_idx = match.player_tables[
            1 - self.position.player_idx
        ].next_character_idx()
        if next_idx is None:
            return super().get_actions(match)
        return [
            self.attack_opposite_active(match, self.damage, self.damage_type),
            SwitchCharacterAction(
                player_idx=1 - self.position.player_idx, character_idx=next_idx
            ),
            self.charge_self(1),
        ]


class DandelionBreeze(ElementalBurstBase):
    name: Literal["Dandelion Breeze"] = "Dandelion Breeze"
    damage: int = -2
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO, elemental_dice_number=4, charge=3
    )
    version: Literal["3.3"] = "3.3"

    def get_actions(self, match: Any) -> List[Actions]:
        ret: List[Actions] = [self.charge_self(-self.cost.charge)]
        characters = match.player_tables[self.position.player_idx].characters
        heal_action = MakeDamageAction(
            damage_value_list=[],
        )
        for character in characters:
            if character.is_alive:
                heal_action.damage_value_list.append(
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=character.position,
                        damage=self.damage,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=self.cost.copy(),
                    )
                )
        ret.append(heal_action)
        ret.append(self.create_summon("Dandelion Field", {"version": self.version}))
        return ret


# Talents


class LandsOfDandelion_3_3(SkillTalent):
    name: Literal["Lands of Dandelion"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Jean"] = "Jean"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO, elemental_dice_number=4, charge=3
    )
    skill: Literal["Dandelion Breeze"] = "Dandelion Breeze"


# character base


class Jean_3_3(CharacterBase):
    name: Literal["Jean"]
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[PhysicalNormalAttackBase | GaleBlade | DandelionBreeze] = []
    faction: List[FactionType] = [FactionType.MONDSTADT]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Favonius Bladework",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            GaleBlade(),
            DandelionBreeze(),
        ]


register_class(Jean_3_3 | DandelionField_3_3 | LandsOfDandelion_3_3)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...modifiable_values import (
    CostValue,
    DamageElementEnhanceValue,
    DamageIncreaseValue,
)
from ...action import Actions, RemoveObjectAction
from ...event import (
    MakeDamageEventArguments,
    RoundPrepareEventArguments,
    SkillEndEventArguments,
)
from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    ELEMENT_TO_ENCHANT_ICON,
    CostLabels,
    DamageElementalType,
    DieColor,
    ElementType,
    ElementalReactionType,
    IconType,
    ObjectPositionType,
    SkillType,
)
from .base import (
    CharacterStatusBase,
    ElementalInfusionCharacterStatus,
    RoundCharacterStatus,
)


class MidareRanzan_3_8(ElementalInfusionCharacterStatus):
    name: Literal[
        # newly created ranzan, and will change its element based on kazuha
        # skill.
        "Midare Ranzan: New",
        # no swirl reaction, anemo element
        "Midare Ranzan",
        # swirl, corresponding element
        "Midare Ranzan: Pyro",
        "Midare Ranzan: Hydro",
        "Midare Ranzan: Cryo",
        "Midare Ranzan: Electro",
    ] = "Midare Ranzan"
    element: ElementType = ElementType.NONE
    version: Literal["3.8"] = "3.8"
    usage: int = 1
    max_usage: int = 1
    newly_created: bool = True
    icon_type: Literal[
        IconType.ELEMENT_ENCHANT_ELEC,
        IconType.ELEMENT_ENCHANT_FIRE,
        IconType.ELEMENT_ENCHANT_WATER,
        IconType.ELEMENT_ENCHANT_ICE,
        IconType.ELEMENT_ENCHANT_WIND,
    ] = IconType.ELEMENT_ENCHANT_WIND

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Do not call ElementalInfusionCharacterStatus.__init__(), because the
        element type is not determined at this time.
        """
        super(ElementalInfusionCharacterStatus, self).__init__(*args, **kwargs)

    def value_modifier_DAMAGE_ELEMENT_ENHANCE(
        self,
        value: DamageElementEnhanceValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageElementEnhanceValue:
        """
        Assert that self.element is correctly set, and if not plunging attack,
        do not modify.
        """
        assert (
            self.newly_created or self.element != ElementType.NONE
        ), "not newly created, but element not set"
        if not match.player_tables[self.position.player_idx].plunge_satisfied:
            # not plunging attack, do nothing
            return value
        return super().value_modifier_DAMAGE_ELEMENT_ENHANCE(value, match, mode)

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        """
        if plunging attack, +1 DMG.
        """
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not self normal attack, do nothing
            return value
        if not match.player_tables[self.position.player_idx].plunge_satisfied:
            # not plunging attack, do nothing
            return value
        # plunging attack, +1 DMG
        value.damage += 1
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        when damage made, if kazuha did elemental skill, change element type
        based on damage results. This should be triggered immediately after
        it is created and by kazuha's elemental skill, and the element type
        is determined at that time.
        """
        if self.name != "Midare Ranzan: New":
            # not newly created, do nothing
            return []
        skill_used: bool = False
        for damage in event.damages:
            if not self.position.check_position_valid(
                damage.final_damage.position,
                match,
                player_idx_same=True,
                character_idx_same=True,
                target_area=ObjectPositionType.SKILL,
            ):  # pragma: no cover
                # not self character use skill, do nothing
                continue
            skill = match.get_object(damage.final_damage.position)
            assert skill.skill_type == SkillType.ELEMENTAL_SKILL
            skill_used = True
            if damage.elemental_reaction != ElementalReactionType.SWIRL:
                # not swirl, do nothing
                continue
            # swirl, change element type
            assert self.name == "Midare Ranzan: New"
            for element in damage.reacted_elements:
                if element != ElementType.ANEMO:
                    self.element = element
            self.name = f"Midare Ranzan: {self.element.name.capitalize()}"  # type: ignore
            self.icon_type = ELEMENT_TO_ENCHANT_ICON[self.element]  # type: ignore
        else:
            assert skill_used, "kazuha elemental skill damage not found"
            if self.element == ElementType.NONE:
                # no swirl, change to anemo
                self.name = "Midare Ranzan"
                self.element = ElementType.ANEMO
        self.infused_elemental_type = ELEMENT_TO_DAMAGE_TYPE[self.element]
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When skill end by this character, if it is newly created, mark False
        and do nothing. Otherwise, remove itself.
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not this character use skill, do nothing
            return []
        if self.newly_created:
            # this card is newly created, mark False and do nothing.
            self.newly_created = False
            return []
        # remove itself
        return [RemoveObjectAction(object_position=self.position)]


class YakshasMask_3_7(ElementalInfusionCharacterStatus, RoundCharacterStatus):
    name: Literal["Yaksha's Mask"] = "Yaksha's Mask"
    version: Literal["3.7"] = "3.7"
    usage: int = 2
    max_usage: int = 2
    infused_elemental_type: DamageElementalType = DamageElementalType.ANEMO
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS
    switch_cost_decrease_usage: int = 1
    switch_cost_decrease_max_usage: int = 1

    skill_cost_decrease_usage: int = 0

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        self.switch_cost_decrease_usage = self.switch_cost_decrease_max_usage
        return super().event_handler_ROUND_PREPARE(event, match)

    def renew(self, new_status: "YakshasMask_3_7") -> None:
        self.switch_cost_decrease_usage = new_status.switch_cost_decrease_usage
        self.skill_cost_decrease_usage = new_status.skill_cost_decrease_usage
        super().renew(new_status)

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not self use damage skill
            return value
        if (
            value.is_corresponding_character_use_damage_skill(
                self.position, match, SkillType.NORMAL_ATTACK
            )
            and match.player_tables[self.position.player_idx].plunge_satisfied
        ):
            # self use plunge attack, increase damage by 2
            value.damage += 2
        if (  # pragma: no branch
            value.damage_elemental_type == DamageElementalType.ANEMO
        ):
            # anemo damage, increase damage by 1
            value.damage += 1
        return value

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        if value.cost.label & CostLabels.SWITCH_CHARACTER.value != 0:
            # switch character cost
            if (
                value.position.player_idx != self.position.player_idx
                or value.position.character_idx != self.position.character_idx
                or self.switch_cost_decrease_usage <= 0
            ):
                # not self switch from this character, or no usage
                return value
            # decrease cost
            if value.cost.decrease_cost(None):  # pragma: no branch
                if mode == "REAL":
                    self.switch_cost_decrease_usage -= 1
        elif value.cost.label & CostLabels.ELEMENTAL_SKILL.value != 0:
            # elemental skill
            if not (
                self.position.check_position_valid(
                    value.position,
                    match,
                    player_idx_same=True,
                    character_idx_same=True,
                    target_area=ObjectPositionType.SKILL,
                )
                and self.skill_cost_decrease_usage > 0
            ):
                # not self use skill, or no usage
                return value
            # decrease cost
            if value.cost.decrease_cost(DieColor.ANEMO):  # pragma: no branch
                if mode == "REAL":
                    self.skill_cost_decrease_usage -= 1
        return value


class Windfavored_4_1(CharacterStatusBase):
    """
    As attack target will be changed, we do not contain triggers for this
    status; instead, maintenance its usage by normal attack. The only exception
    is remove self, when its usage becomes zero in skill end, we will remove
    self.
    """

    name: Literal["Windfavored"] = "Windfavored"
    version: Literal["4.1"] = "4.1"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[Actions]:
        """
        If out of usage, remove self. We do not remove it when usage changes
        because talent card should check whether this status exists. As it is
        triggered after talent card, even its usage becomes zero, talent card
        will still be triggered. Usage change should only happen when this
        character performs normal attack.
        """
        if self.usage <= 0:
            return [RemoveObjectAction(object_position=self.position)]
        return []


register_class(MidareRanzan_3_8 | YakshasMask_3_7 | Windfavored_4_1)



from typing import List, Any, Literal

from .query import satisfy

from ..utils import BaseModel
from .consts import ObjectPositionType, DieColor


class ObjectPosition(BaseModel):
    """
    Position of an object in the game table, which will be set at initializing
    or updated when its position changes. Current change event: card go from
    deck to hand, from hand to deck, from hand to character, from character to
    hand, from character to character, from hand to support, from support to
    support.

    Note the index of the object is not included (n-th summon from player 1's
    sommon lists) as it will change when some action triggered, for example
    RemoveSummonAction. For these actions, we will use the id of the object
    to identify.

    Position is mainly used for objects to decide whether to respond events.

    It is immutable.

    Args:
        player_idx (int): The player index of the object.
        character_idx (int): The character index of the object. If it is not
            on character, set to -1.
        area (ObjectPositionType): The area of the object.
        id (int): The id of the object. If it is not on table, set to -1.
            If it represents SYSTEM, set to 0.
    """

    player_idx: int
    character_idx: int = -1
    area: ObjectPositionType
    id: int

    def __init__(self, *argv, **kwargs) -> None:
        super().__init__(*argv, **kwargs)
        if self.area in [ObjectPositionType.INVALID, ObjectPositionType.SYSTEM]:
            # invalid or system position, do not check other attributes
            return
        # check player_idx is propoerly set
        assert self.player_idx >= 0, "player_idx should be non-negative."
        # check character_idx is properly set
        if self.area in [
            ObjectPositionType.CHARACTER_STATUS,
            ObjectPositionType.SKILL,
            ObjectPositionType.CHARACTER,
        ]:
            assert self.character_idx >= 0, "character_idx should be non-negative."

    def __setattr__(self, name: str, value: Any) -> None:
        """
        Override __setattr__ to make it immutable.
        """
        raise AttributeError("ObjectPosition is immutable.")

    def __delattr__(self, name: str) -> None:
        """
        Override __delattr__ to make it immutable.
        """
        raise AttributeError("ObjectPosition is immutable.")

    def set_id(self, id: int) -> "ObjectPosition":
        """
        Return a new ObjectPosition with the id set.
        """
        return ObjectPosition(
            player_idx=self.player_idx,
            character_idx=self.character_idx,
            area=self.area,
            id=id,
        )

    def set_area(self, area: ObjectPositionType) -> "ObjectPosition":
        """
        Return a new ObjectPosition with the area set.
        """
        return ObjectPosition(
            player_idx=self.player_idx,
            character_idx=self.character_idx,
            area=area,
            id=self.id,
        )

    def copy(self) -> "ObjectPosition":
        """
        As it is a immutable object, it is no need to copy.
        """
        raise AssertionError("ObjectPosition is immutable.")

    def check_position_valid(
        self,
        target_position: "ObjectPosition",
        match: Any,
        player_idx_same: bool | None = None,
        character_idx_same: bool | None = None,
        area_same: bool | None = None,
        id_same: bool | None = None,
        source_area: ObjectPositionType | None = None,
        target_area: ObjectPositionType | None = None,
        source_is_active_character: bool | None = None,
        target_is_active_character: bool | None = None,
    ) -> bool:
        """
        @deprecated This function is deprecated. Use `self.satisfy` instead.

        Based on this position, target position and constraints, give whether
        two position relations fit the constraints.
        """
        if player_idx_same is not None:
            if player_idx_same != (self.player_idx == target_position.player_idx):
                return False
        if character_idx_same is not None:
            if character_idx_same != (
                self.character_idx == target_position.character_idx
            ):
                return False
        if area_same is not None:
            if area_same != (self.area == target_position.area):
                return False
        if id_same is not None:
            if id_same != (self.id == target_position.id):
                return False
        if source_area is not None:
            if self.area != source_area:
                return False
        if target_area is not None:
            if target_position.area != target_area:
                return False
        if source_is_active_character is not None:
            cidx = match.player_tables[self.player_idx].active_character_idx
            return source_is_active_character == (self.character_idx == cidx)
        if target_is_active_character is not None:
            cidx = match.player_tables[target_position.player_idx].active_character_idx
            return target_is_active_character == (target_position.character_idx == cidx)
        return True

    def satisfy(
        self,
        command: str,
        target: "ObjectPosition | None" = None,
        match: Any | None = None,
    ) -> bool:
        """
        Check whether object position satisfies the command.
        Optionally receives target `ObjectPosition` and match. If they are not input,
        related command string will raise error.

        Command formats are:
        - `source / target` to select one object position, or `both` to compare between
            two positions
        - if one position selected: `[pidx=? / cidx=? / area=? / active=(true|false)]`
            to check if position fulfills the situation. for area names, they are case
            insensitive. can use multiple times, and all of them should pass.
        - if two position selected (i.e. `both`), `[(pidx|cidx|area|id)=(same|diff)]`
            to compare two positions are same or not. can use multiple times, and all of
            them should pass.
        - `and` to make multiple checks. All check should pass.

        samples:
        - `source area=support` is this object placed to support
        - `source area=character and target area=skill and both pidx=same cidx=same` is
            this object equipped to a character, and target object is skill of this
            character
        - `both pidx=same and source area=hand and target area=skill` this object in
            hand and target object is this player use skill
        """
        return satisfy(self, command, target, match)

    def not_satisfy(
        self,
        command: str,
        target: "ObjectPosition | None" = None,
        match: Any | None = None,
    ) -> bool:
        """
        Check whether object position not satisfies the command.
        Optionally receives target `ObjectPosition` and match. If they are not input,
        related command string will raise error.

        Command formats are the same as `satisfy`. The result is just the opposite
        of `satisfy`.
        """
        return not self.satisfy(command, target, match)


class MultipleObjectPosition(BaseModel):
    positions: List[ObjectPosition]


class Cost(BaseModel):
    """
    The cost, which is used to define original costs of objects.
    When perform cost, should convert into CostValue.
    """

    label: int = 0
    elemental_dice_number: int = 0
    elemental_dice_color: DieColor | None = None
    same_dice_number: int = 0
    any_dice_number: int = 0
    omni_dice_number: int = 0
    charge: int = 0
    arcane_legend: bool = False
    original_value: Any | None = None

    def copy(self, *argv, **kwargs) -> "Cost":
        """
        Do not support extra args. When perform copy, create new
        instance instead of copy.
        """
        assert len(argv) == 0 and len(kwargs) == 0, "Do not support extra args"
        ori_copy = None
        if self.original_value is not None:  # pragma: no cover
            ori_copy = self.original_value.copy()
        return Cost(
            label=self.label,
            elemental_dice_number=self.elemental_dice_number,
            elemental_dice_color=self.elemental_dice_color,
            same_dice_number=self.same_dice_number,
            any_dice_number=self.any_dice_number,
            omni_dice_number=self.omni_dice_number,
            charge=self.charge,
            arcane_legend=self.arcane_legend,
            original_value=ori_copy,
        )

    @property
    def total_dice_cost(self) -> int:
        """
        Return the total dice cost.
        """
        return (
            self.elemental_dice_number
            + self.same_dice_number
            + self.any_dice_number
            + self.omni_dice_number
        )

    def decrease_cost(self, dice_color: DieColor | None) -> bool:
        """
        Decrease the cost by 1 dice of the given color in-place.

        Return True if decrease successfully, False if not.

        Args:
            dice_color (DieColor | None): The dice color to be decreased.
                If None, decrease same or any dice; Otherwise, decrease
                elemental or any dice.
        """
        assert dice_color != DieColor.OMNI, "Omni dice is not supported yet."
        if dice_color is None:
            # decrease same or any dice
            if self.same_dice_number > 0:
                self.same_dice_number -= 1
                return True
            if self.any_dice_number > 0:
                self.any_dice_number -= 1
                return True
            return False
        # decrease elemental or any dice
        if self.elemental_dice_color == dice_color and self.elemental_dice_number > 0:
            self.elemental_dice_number -= 1
            return True
        if self.any_dice_number > 0:
            self.any_dice_number -= 1
            return True
        return False

    def is_valid(
        self, dice_colors: List[DieColor], charge: int, arcane_legend: bool, strict=True
    ) -> bool:
        """
        Check if dice colors matches the dice cost value.

        Args:
            dice_colors (List[DieColor]): The dice colors to be checked.
            charge (int): The charges to be checked.
            strict (bool): If True, the dice colors must match the dice cost
                value strictly. If False, the dice colors can be more than the
                cost. Note charges always match unstictly.
        """
        # first charge check
        if charge < self.charge:
            return False
        # then arcane legend check
        if self.arcane_legend and not arcane_legend:
            return False

        # then dice number check, no minus
        assert self.elemental_dice_number >= 0
        assert self.same_dice_number >= 0
        assert self.any_dice_number >= 0
        assert self.omni_dice_number >= 0

        # then dice check
        assert self.omni_dice_number == 0, "Omni dice is not supported yet."
        if self.same_dice_number > 0:
            assert (
                self.elemental_dice_number == 0 and self.any_dice_number == 0
            ), "Same dice and elemental/any dice cannot be both used now."
        assert not (
            self.elemental_dice_number > 0 and self.elemental_dice_color is None
        ), "Elemental dice number and color should be both set."
        if strict:
            if len(dice_colors) != (
                self.elemental_dice_number
                + self.same_dice_number
                + self.any_dice_number
                + self.omni_dice_number
            ):
                return False  # dice number not match
        else:
            if len(dice_colors) < (
                self.elemental_dice_number
                + self.same_dice_number
                + self.any_dice_number
                + self.omni_dice_number
            ):
                return False  # dice number not enough
        d = {}
        for color in dice_colors:
            d[color] = d.get(color, 0) + 1
        omni_num = d.get(DieColor.OMNI, 0)
        if self.elemental_dice_number > 0:
            ele_num = d.get(self.elemental_dice_color, 0)
            if ele_num + omni_num < self.elemental_dice_number:
                return False  # elemental dice not enough
        if self.same_dice_number > 0:
            if DieColor.OMNI not in d:
                d[DieColor.OMNI] = 0
            if d[DieColor.OMNI] >= self.same_dice_number:
                return True
            success = False
            for color, same_num in d.items():
                if color == DieColor.OMNI:
                    continue
                if same_num + omni_num >= self.same_dice_number:
                    success = True
                    break
            return success
        return True


class DeckRestriction(BaseModel):
    """
    Deck restriction to add one card into deck
    """

    type: Literal["NONE", "CHARACTER", "FACTION", "ELEMENT"]
    name: str
    number: int



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ..character_base import PhysicalNormalAttackBase

from ...action import Actions

from .maguu_kenki_3_4 import MaguuKenki_3_4 as MK_3_4
from .maguu_kenki_3_4 import BlusteringBlade as BB_3_4
from .maguu_kenki_3_4 import FrostyAssault as FA_3_4
from .maguu_kenki_3_4 import PseudoTenguSweeper


class BlusteringBlade(BB_3_4):
    damage: int = 1

    def get_actions(self, match: Any) -> List[Actions]:
        """
        gather two actions
        """
        return super(BB_3_4, self).get_actions(match) + [
            self.create_summon("Shadowsword: Lone Gale"),
        ]


class FrostyAssault(FA_3_4):
    damage: int = 1

    def get_actions(self, match: Any) -> List[Actions]:
        """
        gather two actions
        """
        return super(FA_3_4, self).get_actions(match) + [
            self.create_summon("Shadowsword: Galloping Frost"),
        ]


class MaguuKenki_3_3(MK_3_4):
    version: Literal["3.3"]
    skills: List[
        PhysicalNormalAttackBase | BlusteringBlade | FrostyAssault | PseudoTenguSweeper
    ] = []

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Ichimonji",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            BlusteringBlade(),
            FrostyAssault(),
            PseudoTenguSweeper(),
        ]


register_class(MaguuKenki_3_3)



from typing import Literal
from pydantic import validator

from ...utils import accept_same_or_higher_version
from ..consts import IconType, ObjectType
from ..object_base import ObjectBase


class StatusBase(ObjectBase):
    """
    Base class of status.
    """

    name: str
    strict_version_validation: bool = False  # default accept higher versions
    version: str
    show_usage: bool = True
    usage: int
    max_usage: int
    type: Literal[ObjectType.CHARACTER_STATUS, ObjectType.TEAM_STATUS]
    renew_type: Literal["ADD", "RESET", "RESET_WITH_MAX"] = "ADD"

    icon_type: IconType

    @validator("version", pre=True)
    def accept_same_or_higher_version(cls, v: str, values):
        return accept_same_or_higher_version(cls, v, values)

    def renew(self, new_status: "StatusBase") -> None:
        """
        Renew the status.
        """
        assert self.renew_type == "ADD", "Other types not tested"
        if self.max_usage < new_status.max_usage:
            self.max_usage = new_status.max_usage
        if self.max_usage <= self.usage:
            # currently over maximum, ignore renew.
            return
        self.usage += new_status.usage
        if self.max_usage < self.usage:
            self.usage = self.max_usage
        return
        # elif self.renew_type == 'RESET':
        #     self.usage = new_status.usage
        # elif self.renew_type == 'RESET_WITH_MAX':
        #     raise NotImplementedError('Not tested part')
        #     self.usage = max(self.usage, new_status.usage)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...modifiable_values import DamageValue

from ...action import (
    ChangeObjectUsageAction,
    ChargeAction,
    CreateDiceAction,
    CreateObjectAction,
    CreateRandomObjectAction,
    DrawCardAction,
    GenerateSwitchCardRequestAction,
    MakeDamageAction,
    RemoveDiceAction,
    SwitchCharacterAction,
)
from ...consts import (
    ELEMENT_TO_DIE_COLOR,
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
)

from ...object_base import EventCardBase
from ...struct import Cost, DeckRestriction, ObjectPosition


class ElementalResonanceCardBase(EventCardBase):
    element: ElementType

    def get_deck_restriction(self) -> DeckRestriction:
        """
        For element resonance cards, should contain at least 2 characters of
        the corresponding element.
        """
        return DeckRestriction(type="ELEMENT", name=self.element.value, number=2)


name_to_element_type = {
    "Elemental Resonance: Woven Flames": ElementType.PYRO,
    "Elemental Resonance: Woven Ice": ElementType.CRYO,
    "Elemental Resonance: Woven Stone": ElementType.GEO,
    "Elemental Resonance: Woven Thunder": ElementType.ELECTRO,
    "Elemental Resonance: Woven Waters": ElementType.HYDRO,
    "Elemental Resonance: Woven Weeds": ElementType.DENDRO,
    "Elemental Resonance: Woven Winds": ElementType.ANEMO,
}


class WovenCards_3_3(ElementalResonanceCardBase):
    name: Literal[
        "Elemental Resonance: Woven Flames",
        "Elemental Resonance: Woven Ice",
        "Elemental Resonance: Woven Stone",
        "Elemental Resonance: Woven Thunder",
        "Elemental Resonance: Woven Waters",
        "Elemental Resonance: Woven Weeds",
        "Elemental Resonance: Woven Winds",
    ]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost()
    element: ElementType = ElementType.NONE  # will update element in __init__

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        self.element = name_to_element_type[self.name]

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateDiceAction]:
        """
        Create one corresponding element dice
        """
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                color=ELEMENT_TO_DIE_COLOR[self.element],
                number=1,
            )
        ]


class ShatteringIce_3_3(ElementalResonanceCardBase):
    name: Literal["Elemental Resonance: Shattering Ice"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=1)
    element: ElementType = ElementType.CRYO

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # active character
        return [
            match.player_tables[self.position.player_idx]
            .get_active_character()
            .position
        ]

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Create status
        """
        assert target is not None
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=target.set_area(ObjectPositionType.CHARACTER_STATUS),
                object_arguments={},
            )
        ]


class SoothingWater_3_3(ElementalResonanceCardBase):
    name: Literal["Elemental Resonance: Soothing Water"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(elemental_dice_color=DieColor.HYDRO, elemental_dice_number=1)
    element: ElementType = ElementType.HYDRO

    def is_valid(self, match: Any) -> bool:
        """
        At least one character is taken damage.
        """
        for character in match.player_tables[self.position.player_idx].characters:
            if character.is_alive and character.damage_taken > 0:
                return True
        return False

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[MakeDamageAction]:
        """
        heal active 2, others 1.
        """
        active_character = match.player_tables[
            self.position.player_idx
        ].get_active_character()
        damage_action = MakeDamageAction(
            damage_value_list=[
                DamageValue(
                    position=self.position,
                    damage_type=DamageType.HEAL,
                    target_position=active_character.position,
                    damage=-2,
                    damage_elemental_type=DamageElementalType.HEAL,
                    cost=self.cost.copy(),
                )
            ],
        )
        for character in match.player_tables[self.position.player_idx].characters:
            if character.is_alive and character.position != active_character.position:
                damage_action.damage_value_list.append(
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=character.position,
                        damage=-1,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=self.cost.copy(),
                    )
                )
        return [damage_action]


class FerventFlames_3_3(ElementalResonanceCardBase):
    name: Literal["Elemental Resonance: Fervent Flames"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=1)
    element: ElementType = ElementType.PYRO

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # active character
        return [
            match.player_tables[self.position.player_idx]
            .get_active_character()
            .position
        ]

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Create status
        """
        assert target is not None
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=target.set_area(ObjectPositionType.CHARACTER_STATUS),
                object_arguments={},
            )
        ]


class HighVoltage_3_3(ElementalResonanceCardBase):
    name: Literal["Elemental Resonance: High Voltage"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=1)
    element: ElementType = ElementType.ELECTRO

    def is_valid(self, match: Any) -> bool:
        """
        At least one character without maximum Energy.
        """
        for character in match.player_tables[self.position.player_idx].characters:
            if character.is_alive and character.charge < character.max_charge:
                return True
        return False

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ChargeAction]:
        """
        active character first to gain energy.
        """
        assert target is None
        active_character = match.player_tables[
            self.position.player_idx
        ].get_active_character()
        if active_character.charge < active_character.max_charge:
            return [
                ChargeAction(
                    player_idx=self.position.player_idx,
                    character_idx=active_character.position.character_idx,
                    charge=1,
                )
            ]
        for character in match.player_tables[self.position.player_idx].characters:
            if character.is_alive and character.charge < character.max_charge:
                return [
                    ChargeAction(
                        player_idx=self.position.player_idx,
                        character_idx=character.position.character_idx,
                        charge=1,
                    )
                ]
        else:
            raise AssertionError("No character can be charged.")


class ImpetuousWinds_3_3(ElementalResonanceCardBase):
    name: Literal["Elemental Resonance: Impetuous Winds"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(elemental_dice_color=DieColor.ANEMO, elemental_dice_number=1)
    element: ElementType = ElementType.ANEMO

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # all alive characters except active character
        active_idx = match.player_tables[self.position.player_idx].active_character_idx
        return [
            character.position
            for (cid, character) in enumerate(
                match.player_tables[self.position.player_idx].characters
            )
            if character.is_alive and cid != active_idx
        ]

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateDiceAction | SwitchCharacterAction]:
        """
        Create one omni element dice
        """
        assert target is not None
        return [
            SwitchCharacterAction(
                player_idx=target.player_idx,
                character_idx=target.character_idx,
            ),
            CreateDiceAction(
                player_idx=self.position.player_idx,
                color=DieColor.OMNI,
                number=1,
            ),
        ]


class EnduringRock_3_3(ElementalResonanceCardBase):
    name: Literal["Elemental Resonance: Enduring Rock"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(elemental_dice_color=DieColor.GEO, elemental_dice_number=1)
    element: ElementType = ElementType.GEO

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Create status
        """
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={},
            )
        ]


class SprawlingGreenery_3_3(ElementalResonanceCardBase):
    name: Literal["Elemental Resonance: Sprawling Greenery"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(elemental_dice_color=DieColor.DENDRO, elemental_dice_number=1)
    element: ElementType = ElementType.DENDRO

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction | ChangeObjectUsageAction]:
        """
        Create status and add usage if exists.
        """
        assert target is None
        ret: List[CreateObjectAction | ChangeObjectUsageAction] = [
            CreateObjectAction(
                object_name=self.name,
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={},
            )
        ]
        object_names = ["Burning Flame", "Dendro Core", "Catalyzing Field"]
        status = match.player_tables[self.position.player_idx].team_status
        for s in status:
            if s.name in object_names:
                ret.append(
                    ChangeObjectUsageAction(
                        object_position=s.position,
                        change_usage=1,
                    )
                )
        summons = match.player_tables[self.position.player_idx].summons
        for s in summons:
            if s.name in object_names:
                ret.append(
                    ChangeObjectUsageAction(
                        object_position=s.position,
                        change_usage=1,
                    )
                )
        return ret


class NationResonanceCardBase(EventCardBase):
    faction: FactionType

    def get_deck_restriction(self) -> DeckRestriction:
        """
        For nation resonance cards, should contain at least 2 characters of
        the corresponding faction.
        """
        return DeckRestriction(type="FACTION", name=self.faction.value, number=2)


class WindAndFreedom_4_1(NationResonanceCardBase):
    name: Literal["Wind and Freedom"]
    version: Literal["4.1"] = "4.1"
    cost: Cost = Cost(same_dice_number=1)
    faction: FactionType = FactionType.MONDSTADT

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Create Wind and Freedom team status.
        """
        assert target is None
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={"version": self.version},
            )
        ]


class WindAndFreedom_3_7(WindAndFreedom_4_1):
    name: Literal["Wind and Freedom"]
    version: Literal["3.7"]


class StoneAndContracts_3_7(NationResonanceCardBase):
    name: Literal["Stone and Contracts"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(any_dice_number=3)
    faction: FactionType = FactionType.LIYUE

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        create Stone and Contracts team status.
        """
        assert target is None
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={"version": self.version},
            )
        ]


class ThunderAndEternity_4_0(NationResonanceCardBase):
    name: Literal["Thunder and Eternity"]
    version: Literal["4.0"] = "4.0"
    cost: Cost = Cost()
    faction: FactionType = FactionType.INAZUMA

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_dice_color(self, match: Any) -> DieColor:
        """
        Return omni color
        """
        return DieColor.OMNI

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[RemoveDiceAction | CreateDiceAction]:
        """
        remove all dice, and create same number dice
        """
        assert target is None
        player_table = match.player_tables[self.position.player_idx]
        dice_number = len(player_table.dice.colors)
        return [
            RemoveDiceAction(
                player_idx=self.position.player_idx,
                dice_idxs=list(range(dice_number)),
            ),
            CreateDiceAction(
                player_idx=self.position.player_idx,
                color=self.get_dice_color(match),
                number=len(player_table.dice.colors),
            ),
        ]


class ThunderAndEternity_3_7(ThunderAndEternity_4_0):
    name: Literal["Thunder and Eternity"]
    version: Literal["3.7"]

    def get_dice_color(self, match: Any) -> str:
        """
        Convert all your Elemental Dice to the Type of the active character.
        """
        character = match.player_tables[self.position.player_idx].get_active_character()
        return ELEMENT_TO_DIE_COLOR[character.element]


class NatureAndWisdom_3_7(NationResonanceCardBase):
    name: Literal["Nature and Wisdom"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=1)
    faction: FactionType = FactionType.SUMERU

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[DrawCardAction | GenerateSwitchCardRequestAction]:
        """
        draw 1 card, and switch any cards in your hand.
        """
        assert target is None
        return [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=1,
                draw_if_filtered_not_enough=True,
            ),
            GenerateSwitchCardRequestAction(
                player_idx=self.position.player_idx,
            ),
        ]


class AbyssalSummons_3_3(NationResonanceCardBase):
    name: Literal["Abyssal Summons"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=2)
    faction: FactionType = FactionType.MONSTER

    def is_valid(self, match: Any) -> bool:
        # if no place for summon, return False
        table = match.player_tables[self.position.player_idx]
        return len(table.summons) < match.config.max_summon_number

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateRandomObjectAction]:
        """
        Create 1 Random Hilichurl Summon!
        """
        assert target is None
        assert not match.config.recreate_mode, "Currently not support recreate mode."
        hilichurl_names = [
            "Cryo Hilichurl Shooter",
            "Electro Hilichurl Shooter",
            "Hilichurl Berserker",
            "Hydro Samachurl",
        ]
        # if already has above summons, do not re-create it
        summons = match.player_tables[self.position.player_idx].summons
        for summon in summons:
            if summon.name in hilichurl_names:
                hilichurl_names.remove(summon.name)
        assert len(hilichurl_names) > 0, "No Hilichurl summon can be created."
        return [
            CreateRandomObjectAction(
                object_names=hilichurl_names,
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.SUMMON,
                    id=-1,
                ),
                object_arguments={},
                number=1,
            )
        ]


class FatuiConspiracy_3_7(NationResonanceCardBase):
    name: Literal["Fatui Conspiracy"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=2)
    faction: FactionType = FactionType.FATUI

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateRandomObjectAction]:
        """
        Create 1 Random Status
        """
        assert target is None
        assert not match.config.recreate_mode, "Currently not support recreate mode."
        fatui_names = [
            "Fatui Ambusher: Cryo Cicin Mage",
            "Fatui Ambusher: Mirror Maiden",
            "Fatui Ambusher: Pyroslinger Bracer",
            "Fatui Ambusher: Electrohammer Vanguard",
        ]
        # if enemy already has above status, do not re-create it
        enemy_status = match.player_tables[1 - self.position.player_idx].team_status
        for status in enemy_status:
            if status.name in fatui_names:
                fatui_names.remove(status.name)
        assert len(fatui_names) > 0, "No Fatui status can be created."
        # create with random name
        return [
            CreateRandomObjectAction(
                object_names=fatui_names,
                object_position=ObjectPosition(
                    player_idx=1 - self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={},
                number=1,
            )
        ]


register_class(
    WovenCards_3_3
    | ShatteringIce_3_3
    | SoothingWater_3_3
    | FerventFlames_3_3
    | HighVoltage_3_3
    | ImpetuousWinds_3_3
    | EnduringRock_3_3
    | SprawlingGreenery_3_3
)
register_class(
    WindAndFreedom_4_1
    | StoneAndContracts_3_7
    | ThunderAndEternity_4_0
    | NatureAndWisdom_3_7
    | AbyssalSummons_3_3
    | FatuiConspiracy_3_7
    | WindAndFreedom_3_7
    | ThunderAndEternity_3_7
)
